@@ -176,47 +176,150 @@ server_description_by_hostname (mongoc_topology_description_t *topology,
  */
 void
 process_sdam_test_ismaster_responses (bson_t *phase,
-                                      mongoc_topology_description_t *td)
+                                      mongoc_topology_t *topology)
 {
+   mongoc_topology_description_t *td;
    mongoc_server_description_t *sd;
-   bson_t ismasters;
-   bson_t ismaster;
-   bson_t response;
    bson_iter_t phase_field_iter;
-   bson_iter_t ismaster_iter;
-   bson_iter_t ismaster_field_iter;
    const char *hostname;
 
+   td = &topology->description;
+   if (bson_iter_init_find (&phase_field_iter, phase, "description")) {
+      const char *description;
+
+      description = bson_iter_utf8 (&phase_field_iter, NULL);
+      MONGOC_DEBUG ("phase: %s", description);
+   }
    /* grab ismaster responses out and feed them to topology */
-   BSON_ASSERT (bson_iter_init_find (&phase_field_iter, phase, "responses"));
-   bson_iter_bson (&phase_field_iter, &ismasters);
-   bson_iter_init (&ismaster_iter, &ismasters);
+   if (bson_iter_init_find (&phase_field_iter, phase, "responses")) {
+      bson_t ismasters;
+      bson_t ismaster;
+      bson_t response;
+      bson_iter_t ismaster_iter;
+      bson_iter_t ismaster_field_iter;
 
-   while (bson_iter_next (&ismaster_iter)) {
-      bson_iter_bson (&ismaster_iter, &ismaster);
+      bson_iter_bson (&phase_field_iter, &ismasters);
+      bson_iter_init (&ismaster_iter, &ismasters);
 
-      bson_iter_init_find (&ismaster_field_iter, &ismaster, "0");
-      hostname = bson_iter_utf8 (&ismaster_field_iter, NULL);
-      sd = server_description_by_hostname (td, hostname);
+      while (bson_iter_next (&ismaster_iter)) {
+         bson_iter_bson (&ismaster_iter, &ismaster);
 
-      /* if server has been removed from topology, skip */
-      if (!sd) {
-         continue;
+         bson_iter_init_find (&ismaster_field_iter, &ismaster, "0");
+         hostname = bson_iter_utf8 (&ismaster_field_iter, NULL);
+         sd = server_description_by_hostname (td, hostname);
+
+         /* if server has been removed from topology, skip */
+         if (!sd) {
+            continue;
+         }
+
+         bson_iter_init_find (&ismaster_field_iter, &ismaster, "1");
+         bson_iter_bson (&ismaster_field_iter, &response);
+
+         /* send ismaster through the topology description's handler */
+         capture_logs (true);
+         mongoc_topology_description_handle_ismaster (
+            td, sd->id, &response, 1, NULL);
+         if (td->servers->items_len == 0) {
+            ASSERT_CAPTURED_LOG ("topology",
+                                 MONGOC_LOG_LEVEL_WARNING,
+                                 "Last server removed from topology");
+         }
+         capture_logs (false);
       }
+   } else if (bson_iter_init_find (
+                 &phase_field_iter, phase, "applicationErrors")) {
+      bson_t app_errors;
+      bson_iter_t app_error_iter;
+      bson_t app_error;
+      bson_iter_t app_error_field_iter;
 
-      bson_iter_init_find (&ismaster_field_iter, &ismaster, "1");
-      bson_iter_bson (&ismaster_field_iter, &response);
+      bson_iter_bson (&phase_field_iter, &app_errors);
+      bson_iter_init (&app_error_iter, &app_errors);
 
-      /* send ismaster through the topology description's handler */
-      capture_logs (true);
-      mongoc_topology_description_handle_ismaster (
-         td, sd->id, &response, 1, NULL);
-      if (td->servers->items_len == 0) {
-         ASSERT_CAPTURED_LOG ("topology",
-                              MONGOC_LOG_LEVEL_WARNING,
-                              "Last server removed from topology");
+      while (bson_iter_next (&app_error_iter)) {
+         uint32_t generation = 0;
+         uint32_t max_wire_version = 0;
+         const char *when_str;
+         bool handshake_complete = false;
+         const char *type_str;
+         _mongoc_sdam_app_error_type_t type = 0;
+         bson_t response;
+         bson_error_t err;
+
+         bson_iter_bson (&app_error_iter, &app_error);
+
+         bson_iter_init_find (&app_error_field_iter, &app_error, "address");
+         hostname = bson_iter_utf8 (&app_error_field_iter, NULL);
+         sd = server_description_by_hostname (td, hostname);
+
+         /* if server has been removed from topology, skip */
+         if (!sd) {
+            continue;
+         }
+
+         if (bson_iter_init_find (
+                &app_error_field_iter, &app_error, "generation")) {
+            BSON_ASSERT (BSON_ITER_HOLDS_INT32 (&app_error_field_iter));
+            generation = bson_iter_int32 (&app_error_field_iter);
+         } else {
+            /* Default to the current generation. */
+            generation = sd->generation;
+         }
+
+         BSON_ASSERT (bson_iter_init_find (
+            &app_error_field_iter, &app_error, "maxWireVersion"));
+         BSON_ASSERT (BSON_ITER_HOLDS_INT32 (&app_error_field_iter));
+         max_wire_version = bson_iter_int32 (&app_error_field_iter);
+
+         BSON_ASSERT (
+            bson_iter_init_find (&app_error_field_iter, &app_error, "when"));
+         BSON_ASSERT (BSON_ITER_HOLDS_UTF8 (&app_error_field_iter));
+         when_str = bson_iter_utf8 (&app_error_field_iter, 0);
+         if (0 == strcmp (when_str, "beforeHandshakeCompletes")) {
+            handshake_complete = false;
+         } else if (0 == strcmp (when_str, "afterHandshakeCompletes")) {
+            handshake_complete = true;
+         } else {
+            test_error ("unexpected 'when' value: %s", when_str);
+         }
+
+         BSON_ASSERT (
+            bson_iter_init_find (&app_error_field_iter, &app_error, "type"));
+         BSON_ASSERT (BSON_ITER_HOLDS_UTF8 (&app_error_field_iter));
+         type_str = bson_iter_utf8 (&app_error_field_iter, 0);
+         if (0 == strcmp (type_str, "command")) {
+            type = MONGOC_SDAM_APP_ERROR_COMMAND;
+         } else if (0 == strcmp (type_str, "network")) {
+            type = MONGOC_SDAM_APP_ERROR_NETWORK;
+         } else if (0 == strcmp (type_str, "timeout")) {
+            type = MONGOC_SDAM_APP_ERROR_TIMEOUT;
+         } else {
+            test_error ("unexpected 'type' value: %s", type_str);
+         }
+
+         if (MONGOC_SDAM_APP_ERROR_COMMAND == type) {
+            BSON_ASSERT (bson_iter_init_find (
+               &app_error_field_iter, &app_error, "response"));
+            BSON_ASSERT (BSON_ITER_HOLDS_DOCUMENT (&app_error_field_iter));
+            bson_iter_bson (&app_error_field_iter, &response);
+         }
+
+         memset (&err, 0, sizeof (bson_error_t));
+         bson_mutex_lock (&topology->mutex);
+         _mongoc_topology_handle_app_error (topology,
+                                            sd->id,
+                                            handshake_complete,
+                                            type,
+                                            &response,
+                                            &err,
+                                            max_wire_version,
+                                            generation);
+         bson_mutex_unlock (&topology->mutex);
       }
-      capture_logs (false);
+   } else {
+      test_error (
+         "test does not have 'responses' or 'applicationErrors' array");
    }
 }
 
@@ -467,9 +570,9 @@ assemble_path (const char *parent_path,
    BSON_ASSERT (path_len + name_len + 1 < MAX_TEST_NAME_LENGTH);
 
    memset (dst, '\0', MAX_TEST_NAME_LENGTH * sizeof (char));
-   strncat (dst, parent_path, path_len);
-   strncat (dst, "/", 1);
-   strncat (dst, child_name, name_len);
+   strcat (dst, parent_path);
+   strcat (dst, "/");
+   strcat (dst, child_name);
 
    for (p = dst; *p; ++p) {
       if (*p == '\\') {
@@ -712,7 +815,7 @@ check_version_info (const bson_t *scenario, bool print_reason)
    return true;
 }
 
-static bool
+bool
 check_scenario_version (const bson_t *scenario)
 {
    /* version info can be nested inside "runOn" array */
@@ -763,6 +866,19 @@ check_test_version (const bson_t *test)
       }
    }
 
+   if (bson_has_field (test, "maxServerVersion")) {
+      s = bson_lookup_utf8 (test, "maxServerVersion");
+      test_version = test_framework_str_to_version (s);
+      server_version = test_framework_get_server_version ();
+      if (server_version >= test_version) {
+         if (test_suite_debug_output ()) {
+            printf ("      SKIP, maxServerVersion %s\n", s);
+            fflush (stdout);
+         }
+         return false;
+      }
+   }
+
    if (bson_has_field (test, "ignore_if_server_version_greater_than")) {
       s = bson_lookup_utf8 (test, "ignore_if_server_version_greater_than");
       /* s is like "3.0", don't skip if server is 3.0.x but skip 3.1+ */
@@ -856,6 +972,28 @@ check_topology_type (const bson_t *test)
    return can_proceed;
 }
 
+static void
+_recreate (mongoc_client_t *client,
+           const char *db_name,
+           const char *collection_name)
+{
+   mongoc_collection_t *collection;
+   mongoc_database_t *db;
+
+   if (!db_name || !collection_name) {
+      return;
+   }
+   collection = mongoc_client_get_collection (client, db_name, collection_name);
+   mongoc_collection_drop (collection, NULL);
+   mongoc_collection_destroy (collection);
+
+   db = mongoc_client_get_database (client, db_name);
+   collection = mongoc_database_create_collection (
+      db, collection_name, NULL /* options */, NULL);
+   mongoc_collection_destroy (collection);
+   mongoc_database_destroy (db);
+}
+
 
 static void
 _insert_data (mongoc_collection_t *collection, bson_t *documents)
@@ -898,7 +1036,7 @@ _insert_data (mongoc_collection_t *collection, bson_t *documents)
 
 
 /* insert the documents in a spec test scenario's "data" array */
-static void
+void
 insert_data (const char *db_name,
              const char *collection_name,
              const bson_t *scenario)
@@ -961,7 +1099,7 @@ DONE:
 }
 
 
-static void
+void
 check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
 {
    bson_t data;
@@ -970,6 +1108,7 @@ check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
    mongoc_cursor_t *cursor;
    bson_t query = BSON_INITIALIZER;
    mongoc_read_prefs_t *prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
+   bson_t opts = BSON_INITIALIZER;
 
    bson_lookup_doc (test, "outcome.collection.data", &data);
    ASSERT (bson_iter_init (&iter, &data));
@@ -983,7 +1122,9 @@ check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
       mongoc_collection_set_read_concern (collection, rc);
    }
 
-   cursor = mongoc_collection_find_with_opts (collection, &query, NULL, prefs);
+   BCON_APPEND (&opts, "sort", "{", "_id", BCON_INT32 (1), "}");
+   cursor = mongoc_collection_find_with_opts (collection, &query, &opts, prefs);
+   bson_destroy (&opts);
 
    while (bson_iter_next (&iter)) {
       bson_t expected_doc;
@@ -1029,7 +1170,7 @@ execute_test (const json_test_config_t *config,
 
    /* Select a primary for testing */
    server_id = mongoc_topology_select_server_id (
-      collection->client->topology, MONGOC_SS_WRITE, NULL, &error);
+      client->topology, MONGOC_SS_WRITE, NULL, &error);
    ASSERT_OR_PRINT (server_id, error);
 
    json_test_ctx_init (&ctx, test, client, db, collection, config);
@@ -1042,8 +1183,7 @@ execute_test (const json_test_config_t *config,
       activate_fail_point (client, server_id, test, "failPoint");
    }
 
-   set_apm_callbacks (&ctx, collection->client);
-
+   set_apm_callbacks (&ctx, client);
    json_test_operations (&ctx, test);
 
    if (config->after_test_cb) {
@@ -1082,7 +1222,9 @@ execute_test (const json_test_config_t *config,
       mongoc_client_destroy (outcome_client);
    }
 
-   mongoc_client_set_apm_callbacks (collection->client, NULL, NULL);
+
+   mongoc_client_set_apm_callbacks (client, NULL, NULL);
+
    json_test_ctx_cleanup (&ctx);
 
    if (!_mongoc_cse_is_enabled (client)) {
@@ -1108,13 +1250,22 @@ activate_fail_point (mongoc_client_t *client,
    bool r;
 
    BSON_ASSERT (client);
-   BSON_ASSERT (server_id);
 
    bson_lookup_doc (test, key, &command);
 
    ASSERT_CMPSTR (_mongoc_get_command_name (&command), "configureFailPoint");
-   r = mongoc_client_command_simple_with_server_id (
-      client, "admin", &command, NULL, server_id, NULL, &error);
+   if (server_id) {
+      r = mongoc_client_command_simple_with_server_id (
+         client, "admin", &command, NULL, server_id, NULL, &error);
+   } else {
+      /* Use default "primary" read preference. */
+      r = mongoc_client_command_simple (client,
+                                        "admin",
+                                        &command,
+                                        NULL /* read prefs */,
+                                        NULL /* reply */,
+                                        &error);
+   }
    ASSERT_OR_PRINT (r, error);
 }
 
@@ -1136,8 +1287,13 @@ deactivate_fail_points (mongoc_client_t *client, uint32_t server_id)
    bool r;
    bson_error_t error;
 
-   sd = mongoc_client_get_server_description (client, server_id);
-   BSON_ASSERT (sd);
+   if (server_id) {
+      sd = mongoc_client_get_server_description (client, server_id);
+      BSON_ASSERT (sd);
+   } else {
+      sd = mongoc_client_select_server (client, false, NULL, &error);
+      ASSERT_OR_PRINT (sd, error);
+   }
 
    if (sd->type == MONGOC_SERVER_RS_PRIMARY &&
        sd->max_wire_version >= WIRE_VERSION_RETRY_WRITES) {
@@ -1146,14 +1302,14 @@ deactivate_fail_points (mongoc_client_t *client, uint32_t server_id)
                    " 'mode': 'off'}");
 
       r = mongoc_client_command_simple_with_server_id (
-         client, "admin", command, NULL, server_id, NULL, &error);
+         client, "admin", command, NULL, sd->id, NULL, &error);
       ASSERT_OR_PRINT (r, error);
 
       command = tmp_bson ("{'configureFailPoint': 'failCommand',"
                           " 'mode': 'off'}");
 
       r = mongoc_client_command_simple_with_server_id (
-         client, "admin", command, NULL, server_id, NULL, &error);
+         client, "admin", command, NULL, sd->id, NULL, &error);
 
       /* ignore error from servers that predate "failCommand" fail point */
       if (!r && !strstr (error.message, "failCommand not found")) {
@@ -1161,19 +1317,30 @@ deactivate_fail_points (mongoc_client_t *client, uint32_t server_id)
       }
    }
 
+   if (sd->max_wire_version >= WIRE_VERSION_4_4) {
+      /* failGetMoreAfterCursorCheckout added in 4.4 */
+      command = tmp_bson ("{'configureFailPoint': "
+                          "'failGetMoreAfterCursorCheckout', 'mode': 'off'}");
+      r = mongoc_client_command_simple_with_server_id (
+         client, "admin", command, NULL, sd->id, NULL, &error);
+      ASSERT_OR_PRINT (r, error);
+   }
+
    mongoc_server_description_destroy (sd);
 }
 
 
-static void
+void
 set_uri_opts_from_bson (mongoc_uri_t *uri, const bson_t *opts)
 {
    bson_iter_t iter;
 
    BSON_ASSERT (bson_iter_init (&iter, opts));
    while (bson_iter_next (&iter)) {
+      const char *key = bson_iter_key (&iter);
+
       /* can't use bson_lookup_write_concern etc. with clientOptions format */
-      if (!strcmp (bson_iter_key (&iter), "w")) {
+      if (!strcmp (key, "w")) {
          mongoc_write_concern_t *wc = mongoc_write_concern_new ();
          if (BSON_ITER_HOLDS_UTF8 (&iter)) {
             mongoc_write_concern_set_wtag (wc, bson_iter_utf8 (&iter, NULL));
@@ -1187,30 +1354,32 @@ set_uri_opts_from_bson (mongoc_uri_t *uri, const bson_t *opts)
 
          mongoc_uri_set_write_concern (uri, wc);
          mongoc_write_concern_destroy (wc);
-      } else if (!strcmp (bson_iter_key (&iter), "readConcernLevel")) {
+      } else if (!strcmp (key, "readConcernLevel")) {
          mongoc_read_concern_t *rc = mongoc_read_concern_new ();
          mongoc_read_concern_set_level (rc, bson_iter_utf8 (&iter, NULL));
          mongoc_uri_set_read_concern (uri, rc);
          mongoc_read_concern_destroy (rc);
-      } else if (!strcmp (bson_iter_key (&iter), "readPreference")) {
+      } else if (!strcmp (key, "readPreference")) {
          mongoc_read_prefs_t *read_prefs = mongoc_read_prefs_new (
             read_mode_from_test (bson_iter_utf8 (&iter, NULL)));
          mongoc_uri_set_read_prefs_t (uri, read_prefs);
          mongoc_read_prefs_destroy (read_prefs);
-      } else if (!strcmp (bson_iter_key (&iter), "retryWrites")) {
-         mongoc_uri_set_option_as_bool (
-            uri, "retryWrites", bson_iter_bool (&iter));
-      } else if (!strcmp (bson_iter_key (&iter), "heartbeatFrequencyMS")) {
-         mongoc_uri_set_option_as_int32 (
-            uri, "heartbeatFrequencyMS", bson_iter_int32 (&iter));
-      } else if (!strcmp (bson_iter_key (&iter), "retryReads")) {
-         mongoc_uri_set_option_as_bool (
-            uri, "retryReads", bson_iter_bool (&iter));
-      } else if (!strcmp (bson_iter_key (&iter), "autoEncryptOpts")) {
+      } else if (!strcmp (key, "autoEncryptOpts")) {
          /* Auto encrypt options are set on constructed client, not in URI. */
+      } else if (!strcmp (key, "writeConcern")) {
+         mongoc_write_concern_t *wc =
+            bson_lookup_write_concern (opts, "writeConcern");
+         mongoc_uri_set_write_concern (uri, wc);
+         mongoc_write_concern_destroy (wc);
+      } else if (mongoc_uri_option_is_bool (key)) {
+         mongoc_uri_set_option_as_bool (uri, key, bson_iter_bool (&iter));
+      } else if (mongoc_uri_option_is_int32 (key)) {
+         mongoc_uri_set_option_as_int32 (uri, key, bson_iter_int32 (&iter));
+      } else if (mongoc_uri_option_is_utf8 (key)) {
+         mongoc_uri_set_option_as_utf8 (uri, key, bson_iter_utf8 (&iter, NULL));
       } else {
          MONGOC_ERROR ("Unsupported clientOptions field \"%s\" in %s",
-                       bson_iter_key (&iter),
+                       key,
                        bson_as_json (opts, NULL));
          abort ();
       }
@@ -1308,9 +1477,9 @@ set_auto_encryption_opts (mongoc_client_t *client, bson_t *test)
       bson_free (coll_name);
    } else {
       /* "If ``autoEncryptOpts`` does not include ``keyVaultNamespace``, default
-       * it to ``admin.datakeys``" */
+       * it to ``keyvault.datakeys``" */
       mongoc_auto_encryption_opts_set_keyvault_namespace (
-         auto_encryption_opts, "admin", "datakeys");
+         auto_encryption_opts, "keyvault", "datakeys");
    }
 
    if (bson_iter_init_find (&iter, &opts, "extra")) {
@@ -1338,36 +1507,14 @@ set_auto_encryption_opts (mongoc_client_t *client, bson_t *test)
    mongoc_auto_encryption_opts_destroy (auto_encryption_opts);
 }
 
-
 static bool
-_should_skip_due_to_server_39704 (const bson_t *test)
+_in_blacklist (const bson_t *test, char **blacklist, uint32_t blacklist_len)
 {
-   const char *desc = bson_lookup_utf8 (test, "description");
-   const char *bad_tests[] = {
-      "only first countDocuments includes readConcern",
-      "only first find includes readConcern",
-      "only first aggregate includes readConcern",
-      "only first distinct includes readConcern",
-      "only first runCommand includes readConcern",
-      "transaction options inherited from defaultTransactionOptions",
-      "startTransaction options override defaults",
-      "defaultTransactionOptions override client options",
-      "readConcern snapshot in startTransaction options",
-      "withTransaction inherits transaction options from "
-      "defaultTransactionOptions",
-      "withTransaction explicit transaction options",
-      "withTransaction explicit transaction options override "
-      "defaultTransactionOptions",
-      "withTransaction explicit transaction options override client options"};
    int i;
+   const char *desc = bson_lookup_utf8 (test, "description");
 
-   /* Only an issue for sharded clusters. */
-   if (!test_framework_is_mongos ()) {
-      return false;
-   }
-
-   for (i = 0; i < sizeof (bad_tests) / sizeof (char *); i++) {
-      if (0 == strcmp (desc, bad_tests[i])) {
+   for (i = 0; i < blacklist_len; i++) {
+      if (0 == strcmp (desc, blacklist[i])) {
          return true;
       }
    }
@@ -1375,6 +1522,19 @@ _should_skip_due_to_server_39704 (const bson_t *test)
    return false;
 }
 
+static bool
+_should_skip_due_to_unsupported_operation (const bson_t *test)
+{
+   char *blacklist[] = {"CreateIndex and dropIndex omits default write concern",
+                        "MapReduce omits default write concern"};
+
+   if (_in_blacklist (test, blacklist, sizeof (blacklist) / sizeof (char *))) {
+      return true;
+   }
+
+   return false;
+}
+
 
 /*
  *-----------------------------------------------------------------------
@@ -1395,8 +1555,8 @@ run_json_general_test (const json_test_config_t *config)
    const bson_t *scenario = config->scenario;
    bson_iter_t scenario_iter;
    bson_iter_t tests_iter;
-   const char *db_name;
-   const char *collection_name;
+   const char *db_name, *db2_name;
+   const char *collection_name, *collection2_name;
 
    ASSERT (scenario);
 
@@ -1411,6 +1571,15 @@ run_json_general_test (const json_test_config_t *config)
                         ? bson_lookup_utf8 (scenario, "collection_name")
                         : "test";
 
+   /* database2_name/collection2_name are optional. */
+   db2_name = bson_has_field (scenario, "database2_name")
+                 ? bson_lookup_utf8 (scenario, "database2_name")
+                 : NULL;
+   collection2_name = bson_has_field (scenario, "collection2_name")
+                         ? bson_lookup_utf8 (scenario, "collection2_name")
+                         : NULL;
+
+
    ASSERT (bson_iter_init_find (&scenario_iter, scenario, "tests"));
    ASSERT (BSON_ITER_HOLDS_ARRAY (&scenario_iter));
    ASSERT (bson_iter_recurse (&scenario_iter, &tests_iter));
@@ -1449,11 +1618,12 @@ run_json_general_test (const json_test_config_t *config)
          continue;
       }
 
-      if (_should_skip_due_to_server_39704 (&test)) {
-         fprintf (stderr,
-                  " - %s SKIPPED, reason: SERVER-39704 causes sharded tests to "
-                  "fail when using readConcern: snapshot\n",
-                  description);
+      if (_should_skip_due_to_unsupported_operation (&test)) {
+         fprintf (
+            stderr,
+            " - %s SKIPPED, reason: test requires operation that libmongoc"
+            " does not provide\n",
+            description);
          continue;
       }
 
@@ -1482,7 +1652,7 @@ run_json_general_test (const json_test_config_t *config)
       }
 
       client = mongoc_client_new_from_uri (uri);
-      mongoc_client_set_error_api (client, 2);
+      mongoc_client_set_error_api (client, MONGOC_ERROR_API_VERSION_2);
       test_framework_set_ssl_opts (client);
       /* reconnect right away, if a fail point causes a disconnect */
       client->topology->min_heartbeat_frequency_msec = 0;
@@ -1508,6 +1678,9 @@ run_json_general_test (const json_test_config_t *config)
       }
 
       set_auto_encryption_opts (client, &test);
+      /* Drop and recreate test database/collection if necessary. */
+      _recreate (client, db_name, collection_name);
+      _recreate (client, db2_name, collection2_name);
       insert_data (db_name, collection_name, scenario);
 
       db = mongoc_client_get_database (client, db_name);
